import {Gamesystem} from "./gamesystems/Gamesystem";
import {ScriptAccount} from "./scriptAccounts/ScriptAccount";
import {Transition} from "./gamesystems/transitions/Transition";
import {State} from "./gamesystems/states/State";
import {ProductGamesystem} from "./gamesystems/ProductGamesystem";
import {SimpleGamesystem} from "./gamesystems/SimpleGamesystem";
import {Character} from "./characters/Character";
import {TemplateType} from "./TemplateType";
import {SimpleTemplateGamesystem} from "./gamesystems/SimpleTemplateGamesystem";

export class GameModel {

  gameModelName: string

  gamesystems: Gamesystem<any, any>[] = [];
  scriptAccounts: ScriptAccount[] = [];
  characters: Character[] = []

  constructor(gameModelName: string) {
    this.gameModelName = gameModelName;
  }

  addGamesystem(gamesystem: Gamesystem<any, any>) {
    if(this.findGamesystem(gamesystem.componentName) == undefined) {
      this.gamesystems.push(gamesystem);
    }

  }

  removeGamesystem(gamesystem : Gamesystem<any, any>) {
    if(gamesystem.parentGamesystem == undefined) {
      this.gamesystems = this.gamesystems.filter(g => g !== gamesystem);
    } else {
      (gamesystem.parentGamesystem as ProductGamesystem).removeChildGamesystem(gamesystem);
    }
  }

  createScriptAccount(scriptAccountName: string) {
    if(scriptAccountName != undefined && scriptAccountName.length > 0) {
      const scriptAccount = new ScriptAccount(scriptAccountName, "");
      const searchedScriptAccount = this.scriptAccounts.find(s => s.componentName === scriptAccount.componentName);
      if(searchedScriptAccount == undefined) {
        this.scriptAccounts.push(scriptAccount);
        return scriptAccount;
      }
    }
    return undefined;
  }

  createGamesystem(gamesystemName: string, parentGamesystemName: string | undefined, templateType: TemplateType) {
    if(gamesystemName != undefined && this.findGamesystem(gamesystemName) == undefined) {
      let simpleGamesystem = new SimpleGamesystem(gamesystemName, "");
      if(templateType == TemplateType.CHARACTER) {
        simpleGamesystem = new SimpleTemplateGamesystem<Character>(gamesystemName, "")
      }


      if(parentGamesystemName != undefined) {
        const parentGamesystem = this.findGamesystem(parentGamesystemName);
        if(parentGamesystem instanceof SimpleGamesystem) {
          const parentProductGamesystem = ProductGamesystem.constructFromSimpleGamesystem(parentGamesystem, this);
          parentProductGamesystem.addChildGamesystem(simpleGamesystem);
          simpleGamesystem.parentGamesystem = parentProductGamesystem;
        } else {
          const productParentGamesystem = parentGamesystem as ProductGamesystem;
          productParentGamesystem.addChildGamesystem(simpleGamesystem);
          simpleGamesystem.parentGamesystem = productParentGamesystem;
        }
      } else {
        this.gamesystems.push(simpleGamesystem);

      }
      return simpleGamesystem;
    }
  }

  createCharacter(characterName: string) {
    const searchedCharacter = this.characters.find(character => character.componentName === characterName);
    if(searchedCharacter == undefined) {
      const character = new Character(characterName, "");
      this.characters.push(character)
      return character
    }
    return undefined
  }

  removeScriptAccount(scriptAccount: ScriptAccount) {
    if(scriptAccount != undefined) {
      this.scriptAccounts = this.scriptAccounts.filter(s => s != scriptAccount);
    }
  }

  removeCharacter(character: Character) {
    if(character != undefined) {
      this.characters = this.characters.filter(c => c.componentName !== character.componentName)
    }
  }

  findGamesystem(gamesystemName: string) {
    const gamesystemQueue : Gamesystem<State<any>, Transition<any>>[] = [];
    this.gamesystems.forEach(gamesystem => gamesystemQueue.push(gamesystem));

    while(gamesystemQueue.length > 0) {
      const currentGamesystem = gamesystemQueue.shift()!;
      if(currentGamesystem.componentName == gamesystemName) {
        return currentGamesystem;
      } else if(currentGamesystem instanceof ProductGamesystem) {
        const currentProductGamesystem = currentGamesystem as ProductGamesystem;
        currentProductGamesystem.innerGamesystems.forEach(gamesystem => gamesystemQueue.push(gamesystem));
      }
    }
  }

  addScriptAccount(scriptAccount: ScriptAccount) {
    this.scriptAccounts.push(scriptAccount);
  }

  generateProductSystemContents() {
    this.gamesystems.forEach(gamesystem => {
      if(gamesystem instanceof ProductGamesystem) {
        gamesystem.generateFromChildsystems();
      }
    })
  }

  listGamesystems(templateType: TemplateType): SimpleTemplateGamesystem<any>[] {
    const gamesystemList:  SimpleTemplateGamesystem<any>[] = []
    const gamesystemQueue: Gamesystem<any, any>[] = this.gamesystems.concat()
    while(gamesystemQueue.length > 0) {
      const currentGamesystem = gamesystemQueue.shift()!;

      if(currentGamesystem.templateType === templateType) {
        gamesystemList.push(currentGamesystem as SimpleTemplateGamesystem<any>)
      }

      if(currentGamesystem instanceof ProductGamesystem) {
        currentGamesystem.innerGamesystems.forEach(innerGamesystem => {
          gamesystemQueue.push(innerGamesystem)
        })
      }
    }

    console.log(gamesystemList)
    return gamesystemList;
  }
}
